defmodule Nulla.Models.User do
  use Ecto.Schema
  import Ecto.Changeset
  import Ecto.Query
  alias Nulla.Repo
  alias Nulla.Snowflake
  alias Nulla.Models.User

  @primary_key {:id, :integer, autogenerate: false}
  schema "users" do
    field :username, :string
    field :domain, :string
    field :email, :string
    field :password, :string
    field :is_moderator, :boolean, default: false
    field :realname, :string
    field :bio, :string
    field :location, :string
    field :birthday, :date
    field :fields, {:array, :map}
    field :tags, {:array, :string}
    field :follow_approval, :boolean, default: false
    field :is_bot, :boolean, default: false
    field :is_discoverable, :boolean, default: true
    field :is_indexable, :boolean, default: true
    field :is_memorial, :boolean, default: false
    field :private_key, :string
    field :public_key, :string
    field :avatar, :string
    field :banner, :string
    field :last_active_at, :utc_datetime

    has_many :user_sessions, Nulla.Models.Session
    has_many :notes, Nulla.Models.Note
    has_many :media_attachments, through: [:notes, :media_attachments]

    timestamps(type: :utc_datetime)
  end

  @doc false
  def changeset(user, attrs) do
    user
    |> cast(attrs, [
      :username,
      :domain,
      :email,
      :password,
      :is_moderator,
      :realname,
      :bio,
      :location,
      :birthday,
      :fields,
      :follow_approval,
      :is_bot,
      :is_discoverable,
      :is_indexable,
      :is_memorial,
      :private_key,
      :public_key,
      :avatar,
      :banner,
      :last_active_at
    ])
    |> validate_required([
      :username,
      :domain,
      :email,
      :password,
      :is_moderator,
      :realname,
      :bio,
      :location,
      :birthday,
      :fields,
      :follow_approval,
      :is_bot,
      :is_discoverable,
      :is_indexable,
      :is_memorial,
      :private_key,
      :public_key,
      :avatar,
      :banner,
      :last_active_at
    ])
  end

  def create_user(attrs) do
    id = Snowflake.next_id()

    %User{}
    |> User.changeset(attrs)
    |> Ecto.Changeset.put_change(:id, id)
    |> Repo.insert()
  end

  def get_user_by_username(username), do: Repo.get_by(User, username: username)

  def get_user_by_username!(username), do: Repo.get_by!(User, username: username)

  def get_total_users_count(domain) do
    Repo.aggregate(from(u in User, where: u.domain == ^domain), :count, :id)
  end

  def get_active_users_count(domain, days) do
    cutoff = DateTime.add(DateTime.utc_now(), -days * 86400, :second)

    from(u in User,
      where: u.domain == ^domain and u.last_active_at > ^cutoff
    )
    |> Repo.aggregate(:count, :id)
  end

  def update_last_active(user) do
    user
    |> Ecto.Changeset.change(last_active_at: DateTime.utc_now())
    |> Repo.update()
  end
end
